% SUR implementation of reduced form narrative VAR
% Code by Thorsten Drautzburg
% This version 12/12/2014
%
% modified by Stephane LHUISSIER and Urszula SZCZERBOWICZ

clear all; clc;

addpath 'library'

%% General Input
mp              = 'unconventional';
vardebt         = 'LOANS';
nlags           = 12;    % number of lags
nvars           = 6;     % number of variables
nexo            = 1;     % constant
ndraws          = 1000;  % number of draws from posterior
imstp           = 37;    % number of steps for impulse response analysis
bupp1           = 95;    % upper error band (percentile)
bupp            = 84;    % upper error band (percentile)
blow            = 16;    % lower error band (percentile)
blow1           = 5;     % lower error band (percentile)
bmed            = 50;    % impulse response percentile
nRow            = 2;    % number of row in Figure
nCol            = 3;    % number of column in Figure
GibbsSkip       = 1;    % GibbsSkip-1 draws are deleted between successive draws to control the autocorrelation
GibbsBurnIn     = ndraws*GibbsSkip/2; % Burn-in phase of Gibbs sampler
rhomax          = 1.000;% maximum eigenvalue
rhomax_attempts = 20;   % Attempts before giving up on the stable eigenvalue 
NO_TREND        = 1;    % -1: quadratic trend, 0: linear trend, 1: no trend
DO_DISPLAY      = 1;    % display characteristics of draw

%% Data
data_file = 'data/LSdata.txt';

switch mp
    case 'conventional'
    i_var_str =  {'ZC1Y', 'ZC10Y', 'UNEMP', 'CPI', 'EBP', vardebt};
    i_var_instr = {'ff4_hf'};
    str_sample_init = '1990-01-01';         % Starting date of the sample (include pre-sample)
    str_sample_end  = '2015-12-01';         % End date of the sample
    case 'unconventional'
    i_var_str =  {'ZC10Y', 'ZC1Y', 'UNEMP', 'CPI', 'EBP', vardebt};
    i_var_instr = {'UMP2'};
    str_sample_init = '1990-01-01';         % Starting date of the sample (include pre-sample)
    str_sample_end  = '2015-12-01';         % End date of the sample   
end

vm_loaddata;

%% Stacking into VAR(1) - depending on the type of trend.
mz=size(ZZ,2);

m_all=size(YY,2);
T=length(YY)-nlags; % number of available time periods, given lag length k

if NO_TREND==1
    ConstantTerms=1;
elseif NO_TREND==-1
    ConstantTerms=3;
else
    ConstantTerms=2;
end;
NoParamsAll=m_all*nlags+ConstantTerms;
X=zeros(T, NoParamsAll);
for j=1:nlags
    X(:, (j-1)*m_all+1:j*m_all)=YY(nlags-(j-1):T+nlags-j, :);
end    

X(:, m_all*nlags+1)=[ones(T,1)];
if NO_TREND==-1
   X(:, end-1:end) =[ cumsum(ones(T, 1)) ,cumsum(ones(T, 1)).^2];
else
    %X(:, end)=cumsum(ones(T, 1));
     X(:, end)=ones(T, 1);
end;

Y=YY(nlags+1:end,:);
Z=ZZ(nlags+1:end,:);
Xy=X;
Xz=ones(T,1); % Could easily generalize


%% Set prior
% For a flat prior, these are all irrelevant
% B_prior=zeros(m_all*nlags+1, m_all); % not implemented here
% N_prior=zeros(m_all*nlags+1, m_all*nlags+1); % not implemented here
S_prior=eye(m_all + mz ); 
nu_prior=0;

nu_T=size(Z,1); % k is already taken care of
nu_posterior=nu_prior+nu_T;


%% Prepare for Gibbs-sampler

A0gibbsinv = zeros(nvars,nvars,ndraws);
Fgibbs     = zeros(nvars*nlags+1,nvars,ndraws);

%
imfgibbs = zeros(imstp, nvars, nvars, ndraws);

%ncoef__n = nvars__n*nlags+nvars*nlags+1;

NoXz=size(Xz,2);

DimV=m_all+mz;

X_Gibbs=sparse([kron(eye(m_all),Xy), zeros(nu_T*m_all,NoXz*mz); zeros(nu_T*mz,(m_all*nlags+ConstantTerms)*m_all), kron(eye(mz),Xz)]);
Y_Gibbs=[reshape(Y,nu_T*m_all,1);reshape(Z,nu_T*mz,1)];

S_posterior=@(By,Bz) (nu_prior/nu_posterior)*S_prior + (T/nu_posterior)*(1/T)*[((Y-Xy*By)'*(Y-Xy*By)), ((Y-Xy*By)'*(Z-Xz*Bz)); ...
                        ((Z-Xz*Bz)'*(Y-Xy*By)), ((Z-Xz*Bz)'*(Z-Xz*Bz))];

% Initialize - should be robust to different options
myBy=(Xy'*Xy)\(Xy'*Y); % OLS
myBz=(Xz'*Xz)\(Xz'*Z); % OLS
NoParamsBeta0=numel(myBy)+numel(myBz);

S_0=S_posterior(myBy,myBz); % OLS
dim_N_XY=(m_all*nlags+ConstantTerms)*m_all+mz*NoXz;
vecB_0=[myBy(:); myBz(:)];
%B_ols=reshape(vecB_0(1:m_all*(m_all*nlags+ConstantTerms),1), m_all*nlags+ConstantTerms, m_all)

% Initialize
mat_draws.B=NaN(ndraws,NoParamsBeta0);
mat_draws.V=NaN(ndraws,DimV^2);

irf_mp = zeros(imstp+nlags,nvars,ndraws);

idx_dim = nvars*nlags;
idx_dim1 = nvars*nlags;
imf_endo_gibbs = zeros(idx_dim1,imstp,ndraws);

z__core_gibbs = zeros(idx_dim,idx_dim,ndraws);
b1_gibbs      = zeros(nvars,1,ndraws);

e_gibbs  = zeros(T,ndraws);
hd_gibbs = zeros(T,nvars*nvars,ndraws);
hdgbs    = zeros(T,nvars,ndraws);

% Initialize
d=0;
while d<ndraws
    d=d+1;

    % Draw from Gibbs-Sampler
    gg=0;
    % Begin loop over draws - only the last one is saved
    while gg<(GibbsSkip+(d==1)*GibbsBurnIn)
        % for the first draw, take account of a burn-in phase
        gg=gg+1;

        % Initialize
        rho=rhomax+2;
        % restrict the posterior to ensure the roots drawn are not too explosive

        U_inv=inv(chol(S_0));
        Xtilde=kron(U_inv' , eye(nu_T)) * X_Gibbs;
        Ytilde=kron(U_inv' , eye(nu_T)) * Y_Gibbs;
        N_XY=(Xtilde'*Ytilde);
        [~,Rx]=qr(Xtilde,0);
        Rx_inv=inv(Rx);

        % Try drawing a not too explosive coefficient matrix
        rr=0;
        while rho>rhomax && rr<rhomax_attempts
            rr=rr+1;
            vecB_1=Rx_inv*mvnrnd(Rx_inv'*N_XY, eye(dim_N_XY))'; 

            Bc=reshape(vecB_1(1:m_all*(m_all*nlags+ConstantTerms),1), m_all*nlags+ConstantTerms, m_all);
            K=[reshape(Bc(1:m_all*nlags, :)', m_all, m_all*nlags); eye((nlags-1)*m_all), zeros((nlags-1)*m_all, m_all)];
            rho=abs(max(eig(K))); 
        end            
        % Regressors of IV equation - here just the means, but code generalizes
        Bz=reshape(vecB_1( (m_all*(m_all*nlags+ConstantTerms)+1):end,1), NoXz, mz);

        try
            DS=chol(inv(S_posterior(Bc,Bz))/nu_posterior);
        catch
            rr=rhomax_attempts+1;  
        end;   
        S_1    = inv(wishrnd(DS'*DS,nu_posterior)); %, DS));
        IV_cov = S_1( m_all+(1:mz),1:m_all);
        
        % Identification of b11 and b12 from the covariance matrix of the VAR
        Sigma_m = S_1(1:m_all,1:m_all);
        Sig11   = Sigma_m(1:mz,1:mz);
        Sig21   = Sigma_m(mz+1:m_all,1:mz);
        Sig22   = Sigma_m(mz+1:m_all,mz+1:m_all);

        b21ib11 = (inv(IV_cov(1:mz,1:mz))*IV_cov(1:mz,mz+1:m_all))';
        
        ZZp     = b21ib11*Sig11*b21ib11'-(Sig21*b21ib11'+b21ib11*Sig21')+Sig22;
        b12b12p = (Sig21- b21ib11*Sig11)'*(ZZp\(Sig21- b21ib11*Sig11));
        b11b11p = Sig11-b12b12p;
        b11     = sqrt(b11b11p);
        b1      = [b11; b21ib11*b11];
        
        if rr<=rhomax_attempts 
            % everything went well
            S_0=S_1;
            vecB_0=vecB_1;

            % Impulse Responses - initial shock: eps(1,1)=1
            irf_mp(nlags+1,:,d) = b1(:,1);
            for jj=2:imstp
                lvars = (irf_mp(nlags+jj-1:-1:jj,:,d))';
                irf_mp(nlags+jj,:,d) = lvars(:)'*Bc(1:nlags*m_all,:);   
            end
            
            
            A1 = diag(ones(nvars*(nlags-1),1));
            A2 = [A1 zeros(nvars*(nlags-1),nvars)];
            z__core  = [Bc(1:nvars*nlags,:)'; A2];
            z__core_gibbs(:,:,d) = z__core;
            b1_gibbs(:,:,d) = b1; % impact
    
        else
            % found a covariance matrix that was not PD
            gg=gg-1;
        end;
        
        if DO_DISPLAY 
            fprintf('%1.0f :: ', [d gg rr]);
            fprintf('%1.3f \n ', rho);
        end;
    end;
    % end loop over draws - only the last one is saved
    
    % Save coefficients and variance-covariance matrix
    mat_draws.B(d,:)=[Bc(:)' Bz(:)'];
    mat_draws.V(d,:)=S_1(:)';
    
    %
    Fgibbs(:,:,d) = Bc;
    
    F        = Fgibbs(:,:,d);
    A0gibbsinv(1,:,d) = b1_gibbs(:,:,d)';
    A0hatinv = A0gibbsinv(:,:,d);
    
    imfhat  = fn_impulse(F,A0hatinv,[nvars nlags imstp]);
    imfgibbs(:,:,:,d) = reshape(imfhat,imstp,nvars,nvars);
    
end

%% Impulse responses functions (IRFs)
t = 0:imstp-1;

irf_mp = squeeze(imfgibbs(:,:,1,:));

% memory allocation
irf_mp_quan = zeros(imstp,nvars,5);

% sort the IRFs
sort_irf_mp = sort(irf_mp,3);

% computing mean, upper and lower bands for IRFs 
irf_mp_quan(:,:,1) = sort_irf_mp(:,:,ndraws*blow1/100);
irf_mp_quan(:,:,2) = sort_irf_mp(:,:,ndraws*blow/100);
irf_mp_quan(:,:,3) = sort_irf_mp(:,:,ndraws*bmed/100);
irf_mp_quan(:,:,4) = sort_irf_mp(:,:,ndraws*bupp/100);
irf_mp_quan(:,:,5) = sort_irf_mp(:,:,ndraws*bupp1/100);

figure
for g=1:nvars
    subplot(nRow,nCol,g)
    plot(t,-irf_mp_quan(:,g,5),':','color',[0 0.4470 0.7410],'LineWidth',1.0);
    hold on
    plot(t,-irf_mp_quan(:,g,4),'-.','color',[0 0.4470 0.7410],'LineWidth',1.0);
    hold on
    plot(t,-irf_mp_quan(:,g,3),'-','color',[0 0.4470 0.7410],'LineWidth',1.0);
    hold on
    plot(t,-irf_mp_quan(:,g,2),'-.','color',[0 0.4470 0.7410],'LineWidth',1.0);
    hold on
    plot(t,-irf_mp_quan(:,g,1),':','color',[0 0.4470 0.7410],'LineWidth',1.0);
    xlim([t(1) t(imstp)])
    hold on
    plot(t,zeros(imstp,1),'k:', 'LineWidth', 2.5)
    title(i_var_str{g})
    set(gca,'Layer','top')
    xlabel('\fontsize{8}{months after shock}')
    box on
    grid on
    set(gca,'xtick',[0 6 12 24 36])
    set(gca,'xgrid','on')
    set(gca,'ygrid','on')
end
